﻿// This source is subject to Microsoft Public License (Ms-PL).
// Please see http://simplehidlibrary.codeplex.com/ for details.
// All other rights reserved.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;
using System.Diagnostics;

namespace SimpleHID
{
  /// <summary>
  /// USB HID device
  /// </summary>
  public class HIDDevice : IDisposable
  {
    #region Private stuff

    private IntPtr handle;
    private FileStream fileStream;
    private NativeMethods.HidCaps hidCaps;
    private IntPtr preparedData;
    NativeMethods.HidP_Button_Caps[] inputButtonCaps;
    private bool bFileOpened = false;
    #endregion

    #region Constructor

    /// <summary>
    /// ctor
    /// </summary>
    public HIDDevice()
    {
    }

    #endregion

    #region Init

    /// <summary>
    /// Connect to device
    /// </summary>
    public bool Init(string devicePath)
    {
      handle = NativeMethods.CreateFile(devicePath, NativeMethods.GENERIC_READ | NativeMethods.GENERIC_WRITE, NativeMethods.FILE_SHARE_READ | NativeMethods.FILE_SHARE_WRITE, IntPtr.Zero, NativeMethods.OPEN_EXISTING, NativeMethods.FILE_FLAG_OVERLAPPED, IntPtr.Zero);

      if (handle != NativeMethods.InvalidHandleValue)
      {
          //preparedData = Marshal.AllocHGlobal(Marshal.SizeOf(804));

          if (NativeMethods.HidD_GetPreparsedData(handle, out preparedData))
          {
              try
              {
                  Int32 nHidCapSize = Marshal.SizeOf(typeof(NativeMethods.HidCaps));
                  IntPtr inptrHidCaps = Marshal.AllocHGlobal(nHidCapSize);

                  if (NativeMethods.HidP_GetCaps(preparedData, inptrHidCaps) != NativeMethods.HIDP_STATUS_SUCCESS)
                  {
                      LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]HidP_GetCaps is not HIDP_STATUS_SUCCESS , HidP_GetCaps Failed"), LogLib.SUB_LOG_TYPE.ERROR);
                      return false;
                  }

                  hidCaps = (NativeMethods.HidCaps)Marshal.PtrToStructure(inptrHidCaps, typeof(NativeMethods.HidCaps));

                  var numberInputButtonCaps = hidCaps.NumberInputButtonCaps;
                  Int32 nSize = Marshal.SizeOf(typeof(NativeMethods.HidP_Button_Caps));

                  IntPtr ptrinputButtonCaps = Marshal.AllocHGlobal(nSize * numberInputButtonCaps);

                  if (NativeMethods.HidP_GetButtonCaps(NativeMethods.HIDP_REPORT_TYPE.HidP_Input, ptrinputButtonCaps, ref numberInputButtonCaps, preparedData) != NativeMethods.HIDP_STATUS_SUCCESS)
                  {
                      LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]HidP_GetButtonCaps is not HIDP_STATUS_SUCCESS , HidP_GetButtonCaps Failed"), LogLib.SUB_LOG_TYPE.ERROR);
                      return false;
                  }

                  inputButtonCaps = new NativeMethods.HidP_Button_Caps[numberInputButtonCaps];
                  for (int i = 0; i < numberInputButtonCaps; i++)
                  {
                      IntPtr tmpPtr = new IntPtr((ptrinputButtonCaps.ToInt32() + i * nSize));
                      inputButtonCaps[i] = (NativeMethods.HidP_Button_Caps)Marshal.PtrToStructure(tmpPtr, typeof(NativeMethods.HidP_Button_Caps));
                  }

                  fileStream = new FileStream(new SafeFileHandle(handle, false), FileAccess.Read | FileAccess.Write, hidCaps.InputReportByteLength, true);
                  bFileOpened = true;
                  Marshal.FreeHGlobal(ptrinputButtonCaps);

                  Marshal.FreeHGlobal(inptrHidCaps);
              }
              catch (Exception e)
              {
                  LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]Init Exception : {0}", e.Message.ToString()), LogLib.SUB_LOG_TYPE.ERROR);
                  return false;
              }
          }
          else
          {
              LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]HidD_GetPreparsedData is false, HidD_GetPreparsedData Failed"), LogLib.SUB_LOG_TYPE.ERROR);
              return false;
          }
      }
      else
      {
          LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]CreateFile Failed"), LogLib.SUB_LOG_TYPE.ERROR);
          handle = IntPtr.Zero;
          return false;
      }
      
      return true;
    }

    private IntPtr MarshalToPointer(object data)
    {
        IntPtr buf = Marshal.AllocHGlobal(Marshal.SizeOf(data));
        Marshal.StructureToPtr(data,buf, false);
        return buf;
    }

    #endregion

    #region Input

    /// <summary>
    /// Create raw data for an Input Report with a specific Id
    /// </summary>
    /// <param name="reportId">InputReport Id</param>
    /// <returns>Raw data</returns>
    private byte[] CreateRawInputReport(byte reportId)
    {
      var data = new byte[hidCaps.InputReportByteLength];
      if (NativeMethods.HidP_InitializeReportForID(NativeMethods.HIDP_REPORT_TYPE.HidP_Input, reportId, preparedData, data, data.Length) != NativeMethods.HIDP_STATUS_SUCCESS)
      {
          throw new Win32Exception();
      }
      return data;
    }

    /// <summary>
    /// 
    /// </summary>
    private byte[] ReadRawInputReport()
    {
      var firstInputReportId = inputButtonCaps[0].ReportID;
      var data = CreateRawInputReport(firstInputReportId);
      try
      {
         fileStream.Read(data, 0, data.Length);
      }
      catch (Exception)
      {
        throw new HIDDeviceException();
      }
      return data;
    }

    /// <summary>
    /// 
    /// </summary>
    public byte[] ReadInputReport()
    {
      var data = ReadRawInputReport();
      int length = 100;
      var uu = new NativeMethods.HIDP_DATA[length];
      if (NativeMethods.HidP_GetData(NativeMethods.HIDP_REPORT_TYPE.HidP_Input, uu, ref length, preparedData, data,data.Length) != NativeMethods.HIDP_STATUS_SUCCESS)
      {
        throw new Win32Exception();
      }

      return data;
    }


    public byte[] ReadInputReportAsync()
    {
        var data = ReadRawInputReportAsync();
        int length = 100;
        var uu = new NativeMethods.HIDP_DATA[length];
        if (NativeMethods.HidP_GetData(NativeMethods.HIDP_REPORT_TYPE.HidP_Input, uu, ref length, preparedData, data, data.Length) != NativeMethods.HIDP_STATUS_SUCCESS)
        {
            throw new Win32Exception();
        }

        return data;
    }

    public bool bExit { get; set; }
    private byte[] ReadRawInputReportAsync()
    {
        var firstInputReportId = inputButtonCaps[0].ReportID;
        var data = CreateRawInputReport(firstInputReportId);
        int nByteRead = 0;
        try
        {
            IAsyncResult result = fileStream.BeginRead(data, 0, data.Length, null, null);
            while (!result.IsCompleted && !bExit)
            {
                System.Threading.Thread.Sleep(50);
            }
            
            if(result.IsCompleted == true)
               nByteRead = fileStream.EndRead(result);
        }
        catch (Exception e)
        {
            throw new HIDDeviceException();
        }
        return data;
    }

    public void BeginAsyncRead()
    {
        try
        {
            byte[] arrInputReport = new byte[hidCaps.InputReportByteLength];
            if (bFileOpened == true)
            {
                fileStream.BeginRead(arrInputReport, 0, hidCaps.InputReportByteLength, new AsyncCallback(ReadCompleted), arrInputReport);
            }
        }
        catch (Exception e)
        {
            LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]BeginAsyncRead : {0}", e.Message.ToString()), LogLib.SUB_LOG_TYPE.ERROR);
        }
    }

    protected void ReadCompleted(IAsyncResult iResult)
    {
        byte[] arrBuff = (byte[])iResult.AsyncState;
        try
        {
            if (fileStream != null)
            {
                int nReadCount = fileStream.EndRead(iResult);

                if (RecvUSBInputMsgEvtHandler != null)
                {
                    RecvUSBInputMsgEvtHandler.Invoke((object)arrBuff, EventArgs.Empty);
                }

                System.Threading.Thread.Sleep(5);
            }
        }
        catch (IOException ex)
        {
            LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]IO Exception : {0}", ex.Message.ToString()), LogLib.SUB_LOG_TYPE.ERROR);
        }
        catch (Exception e)
        {
            LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]Exception : {0}", e.Message.ToString()), LogLib.SUB_LOG_TYPE.ERROR);
        }
        finally
        {
            if(fileStream != null)
                BeginAsyncRead();
        }
    }

    public EventHandler RecvUSBInputMsgEvtHandler;
              

    #endregion

    #region Output

//    /// <summary>
//    /// Create raw data for an Output Report with a specific Id
//    /// </summary>
//    /// <param name="reportId">OutputReport Id</param>
//    /// <returns>Raw data</returns>
//    private byte[] CreateRawOutputReport(byte reportId)
//    {
//      var data = new byte[hidCaps.OutputReportByteLength];
//      if (NativeMethods.HidP_InitializeReportForID(NativeMethods.HIDP_REPORT_TYPE.HidP_Output, reportId, preparedData, data, data.Length) != NativeMethods.HIDP_STATUS_SUCCESS)
//      {
//        throw new Win32Exception();
//      }
//      return data;
//    }

//    /// <summary>
//    /// Write binary data
//    /// </summary>
//    private void WriteRawOutputReport(byte[] data)
//    {
////      if (NativeMethods.HidD_SetOutputReport(handle, data, data.Length) == false)
////      {
////        throw new HIDDeviceException();
////      }

//      try
//      {
//        fileStream.Write(data, 0, data.Length);
//        fileStream.Flush();
//      }
//      catch (Exception)
//      {
//        throw new HIDDeviceException();
//      }
//    }

    #endregion

    #region IDisposable implementation

    /// <summary>
    /// Releasing Unmanaged Resources
    /// </summary>
    public void Dispose()
    {
        if (fileStream != null)
        {
            bFileOpened = false;
            fileStream.Close();
            fileStream = null;
        }

        try
        {
            if (preparedData != IntPtr.Zero)
            {
                NativeMethods.HidD_FreePreparsedData(ref preparedData);
            }
        }
        catch (Exception e)
        {
            LogLib.LogManager.OutputDebugString(1, this, string.Format("[HIDDevice]Exception HidD_FreePreparsedData error: {0}", e.Message.ToString()), LogLib.SUB_LOG_TYPE.ERROR);
        }

        if (handle != IntPtr.Zero)
            NativeMethods.CloseHandle(handle);
    }

    #endregion
  }
}