﻿// This source is subject to Microsoft Public License (Ms-PL).
// Please see http://simplehidlibrary.codeplex.com/ for details.
// All other rights reserved.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Text;

namespace SimpleHID
{
  /// <summary>
  /// USB HID helper class
  /// </summary>
  public class HIDManager
  {
    #region Public static methods

    /// <summary>
    /// Return a list information of connected USB HID devices
    /// </summary>
    static public IEnumerable<HIDInfoSet> GetInfoSets()
    {
      var devicePathList = new List<HIDInfoSet>();
      Guid gHid;
      Int32 nGuidSize = Marshal.SizeOf(typeof(Guid));
      IntPtr inptrgHid = Marshal.AllocHGlobal(nGuidSize);

      NativeMethods.HidD_GetHidGuid(inptrgHid);
      gHid = (Guid)Marshal.PtrToStructure(inptrgHid, typeof(Guid));
      var hInfoSet = NativeMethods.SetupDiGetClassDevs(inptrgHid, null, IntPtr.Zero, NativeMethods.DIGCF_DEVICEINTERFACE | NativeMethods.DIGCF_PRESENT);
      if (hInfoSet == NativeMethods.InvalidHandleValue)
      {
         LogLib.LogManager.OutputDebugString(1, null, string.Format("[HIDManager]SetupDiGetClassDevs is InvalidHandleValue"), LogLib.SUB_LOG_TYPE.ERROR);
         return devicePathList;
      }

      NativeMethods.DeviceInterfaceData oInterface = new NativeMethods.DeviceInterfaceData();
      oInterface.Size = Marshal.SizeOf(oInterface);
      int nIndex = 0;

      IntPtr ptrInterFace = Marshal.AllocHGlobal(Marshal.SizeOf(typeof(NativeMethods.DeviceInterfaceData)));
      Marshal.StructureToPtr(oInterface, ptrInterFace, false);
 
      while (NativeMethods.SetupDiEnumDeviceInterfaces(hInfoSet, 0, inptrgHid, (uint)nIndex, ptrInterFace))        // oInterface fixed 필요
      {
        oInterface = (NativeMethods.DeviceInterfaceData)Marshal.PtrToStructure(ptrInterFace, typeof(NativeMethods.DeviceInterfaceData));
        var strDevicePath = GetDevicePath(hInfoSet, ptrInterFace);
        var handle = Open(strDevicePath);
        if (handle != NativeMethods.InvalidHandleValue)
        {
          var productString = GetName(handle);
          var serialNumberString = GetSerialNumber(handle);
          NativeMethods.HIDD_ATTRIBUTES attributes;
          GetAttr(handle, out attributes);
          devicePathList.Add(new HIDInfoSet(productString, serialNumberString, strDevicePath, attributes.VendorID, attributes.ProductID, attributes.VersionNumber));
        }
        nIndex++;	
      }
      if (NativeMethods.SetupDiDestroyDeviceInfoList(hInfoSet) == false)
      {
          LogLib.LogManager.OutputDebugString(1, null, string.Format("[HIDManager]SetupDiDestroyDeviceInfoList is false , SetupDiDestroyDeviceInfoList Failed"), LogLib.SUB_LOG_TYPE.ERROR);
      }

      Marshal.FreeHGlobal(ptrInterFace);
      Marshal.FreeHGlobal(inptrgHid);
      return devicePathList;
    }

    #endregion

    #region Private helper methods

    private static IntPtr Open(string devicePath)
    {
      return NativeMethods.CreateFile(devicePath, NativeMethods.GENERIC_READ | NativeMethods.GENERIC_WRITE, NativeMethods.FILE_SHARE_READ | NativeMethods.FILE_SHARE_WRITE, IntPtr.Zero, NativeMethods.OPEN_EXISTING, NativeMethods.FILE_FLAG_OVERLAPPED, IntPtr.Zero);
    }

    private static void GetAttr(IntPtr handle, out NativeMethods.HIDD_ATTRIBUTES attributes)
    {
        int nSize = Marshal.SizeOf(typeof(NativeMethods.HIDD_ATTRIBUTES));
        IntPtr ptrAttributes = Marshal.AllocHGlobal(nSize);
        if (NativeMethods.HidD_GetAttributes(handle, ptrAttributes) == false)
        {
            LogLib.LogManager.OutputDebugString(1, null, string.Format("[HIDManager]HidD_GetAttributes is false , Can not find hdd Attributes"), LogLib.SUB_LOG_TYPE.ERROR);
        }

        attributes = (NativeMethods.HIDD_ATTRIBUTES)Marshal.PtrToStructure(ptrAttributes, typeof(NativeMethods.HIDD_ATTRIBUTES));
        attributes.Size = Marshal.SizeOf(attributes);
    }

    private static string GetName(IntPtr handle)
    { 
      IntPtr IntPtrProductName = Marshal.AllocHGlobal(1000);

      if (NativeMethods.HidD_GetProductString(handle, IntPtrProductName, 1000) == false)
      {
        LogLib.LogManager.OutputDebugString(1, null, string.Format("[HIDManager]HidD_GetProductString is false , Can not find HidD_GetProductString"), LogLib.SUB_LOG_TYPE.ERROR);
        return "";
      }
      string strProductName = Marshal.PtrToStringAuto(IntPtrProductName);
      Marshal.FreeHGlobal(IntPtrProductName);
      return strProductName;
    }

    private static string GetSerialNumber(IntPtr handle)
    {
      IntPtr IntSerialNumber = Marshal.AllocHGlobal(300);

      if (NativeMethods.HidD_GetSerialNumberString(handle, IntSerialNumber, 300) == false)
      {
        // Some devices has no serial number
        // Just return an empty string
        return string.Empty;
      }
      string strSerialNumber = Marshal.PtrToStringAuto(IntSerialNumber);
      Marshal.FreeHGlobal(IntSerialNumber);
      return strSerialNumber;
    }

    /// <summary>
    /// Get USB HID device path
    /// </summary>
    private static string GetDevicePath(IntPtr hInfoSet, IntPtr oInterface)
    {
        IntPtr ptrInterFace1 = Marshal.AllocHGlobal(Marshal.SizeOf(typeof(NativeMethods.DeviceInterfaceDetailData)));

        NativeMethods.DeviceInterfaceDetailData detIface = new NativeMethods.DeviceInterfaceDetailData();
        detIface.Size = Marshal.SizeOf(typeof(IntPtr)) == 8 ? 8 : 5;

        /* required size */
        uint reqSize = (uint)Marshal.SizeOf(detIface);
        Marshal.StructureToPtr(detIface, ptrInterFace1, false);

        if (NativeMethods.SetupDiGetDeviceInterfaceDetail(hInfoSet, oInterface, ptrInterFace1, reqSize, ref reqSize, IntPtr.Zero) == false)
        {
            // TODO: Find a solution
            // Tip: http://stackoverflow.com/questions/1054748/setupdigetdeviceinterfacedetail-unexplainable-error
            //throw new HIDDeviceException();
        }

        NativeMethods.DeviceInterfaceDetailData DevInterfaceDetailData = (NativeMethods.DeviceInterfaceDetailData)Marshal.PtrToStructure(ptrInterFace1, typeof(NativeMethods.DeviceInterfaceDetailData));
        string DevPath = DevInterfaceDetailData.DevicePath;
        Marshal.FreeHGlobal(ptrInterFace1);
        return DevPath;
    }

    #endregion
  }
}